; Title:		Dock
;				Engine that makes you dock arbitrary top level windows (dock clients) to arbitrary top level window (dock host)
;
; By majkinetor 
;----------------------------------------------------------------------------------------------------------------------------------
; Function:		Dock
;				Instantiate dock on given client upon host. Multiple clients per one host are supported.
;
; Parameters:
;				Dock_HostID	- Global, sets docking host
;	   	   Dock_OnHostDeath - Global, sets label that will be called when host dies. Afterwards global variable
;							  Dock_HostDiead will be true. This routine will be called only once when host dies. If you change
;							  Dock_HostId afterwards you will have to set Dock_HostDied=0. You will be notified again when current 
;							  host dies.
;
;				pClientId	- HWND of the Client GUI
;				pDockDef	- Dock definition, see bellow. To remove dock client pass "-"
;
; Defintion:	xhw,xw,xd, yhh,yh,yd, whw,wd, hhh,hd
;
;				o The x coordinate of the top, left corner of the client window is computed as 
;				HostX + xhw*HostWidth + xw*ClientWidth + xd, with the parameters xhw, xw and xd 
;
;				o The y coordinate of the top, left corner of the client window is computed as 
;				HostY + yhh*HostHeight + yh*ClientHeight + yd, with the parameters yhh, yh and yd 
;
;				o The width ClientWidth of the client window is computed as whw*HostWidth + wd, with the parameters whw and wd. 
;
;				o The height ClientHeight of the client window is computed as hhh*HostHeight + hd, with the parameters hhh and hd.
; Example:
;>	 	Dock(Dock_c1, "0,-1,-10, 0,0,0, 0,63, 1,0")		  ;top left, host height
;>		Dock(Dock_c2, "0,0,0, 0,-1,-5, 1,0,0,30")       ;top above, host width
;
Dock(pClientID, pDockDef) {
	local cnt
	static hwnd, init=0, idDel

	If !(init++) {
		Dock_HostDied := Dock_aClient[0] := 0

	 	hwnd := WinExist("ahk_pid " . DllCall("GetCurrentProcessId"))
		Dock_hookDll = dll\wineventhookq.dll
	
		Dock_hDock := API_LoadLibrary(Dock_hookDll)
		if !Dock_hDock
			return "LoadLibrary failed"
	    Dock_msg := 0x550
		
		SetWinDelay -1

		Dock_hHook1 := Hook(hwnd, Dock_msg, 3)										; EVENT_SYSTEM_FOREGROUND
		Dock_hHook2 := Hook(hwnd, Dock_msg, 0x800B, 0x800B, "Dock_HookHandler")		; EVENT_OBJECT_LOCATIONCHANGE 0x0000800b

		if !(Dock_hHook1 && Dock_hHook2) {
			API_FreeLibrary( Dock_hookDll )		
			return "Hook failed"
		}
	}
	cnt := Dock_aClient[0]

	;remove dock client ?
	if (pDockDef="-") and (Dock_%pClientID%){

		idDel := Dock_%pClientID%

		loop, 10
			Dock_%pClientID%_%A_Index% := 0
		Dock_%pClientID% := ""
		
		;move last one to the place of the deleted one
		Dock_aClient[%idDel%] := Dock_aClient[%cnt%]
		Dock_aClient[%cnt%] := 
		Dock_aClient[0]--
		return
	}

	;add new dock client
	Dock_%pClientID% := ++cnt	
	loop, parse, pDockDef, `,,%A_Space%%A_Tab%
		Dock_%pClientID%_%A_Index% := A_LoopField


	Dock_aClient[%cnt%] := pClientId
	Dock_aClient[0]++

	Dock_Update()

	return "OK"
	
}

;-----------------------------------------------------------------------------------------------------
;Function:	Dock_Update
;		 	You can call this function to update dock manualy 
;
Dock_Update(event=3) {
	local hX, hY, hW, hh, W, H, X, Y
	local cId
	
	WinGetPos hX, hY, hW, hH, ahk_id %Dock_HostID%

	;xhw,xw,xd, yhh,yh,yd, whw,wd, hhh,hd
	loop, % Dock_aClient[0]
	{
		cId := Dock_aClient[%A_Index%]

		W := Dock_%cId%_7*hW + Dock_%cId%_8,  H := Dock_%cId%_9*hH + Dock_%cId%_10
		X := hX + Dock_%cId%_1*hW + Dock_%cId%_2*W + Dock_%cId%_3
		Y := hY + Dock_%cId%_4*hH + Dock_%cId%_5*H + Dock_%cId%_6

		WinMove ahk_id %cId%,,X,Y	;,W,H	removed in LilBuilder

		If (event = 3) {
			WinSet AlwaysOnTop, On,  ahk_id %cId%
;			Sleep 0
			WinSet AlwaysOnTop, Off, ahk_id %cId%
		}
	}
}

;==================================== INTERNAL ======================================================

Dock_HookHandler(wParam, lParam) {
	local event, hwnd

	DllCall("RtlMoveMemory", "UIntP", event , "UInt",lParam+4, "UInt", 4)
	DllCall("RtlMoveMemory", "UIntP", hwnd  , "UInt",lParam+8, "UInt", 4)

	If (hwnd != Dock_HostID){
		if !WinExist("ahk_id " Dock_HostID) && IsLabel(Dock_OnHostDeath) && !Dock_HostDied++
			gosub %Dock_OnHostDeath%
		return
	}

	Dock_Update(event)
}

Hook(comm_hwnd, comm_msg, s_event, e_event="", function="", wparam=0) {
   Global Dock_hookDll
   r := DllCall(Dock_hookDll "\reghook", "UInt" ,comm_hwnd, "UInt", COMM_MSG, "UInt", s_event, "UInt", e_event ? e_event : s_event, "UInt",wparam)
   If (function)
      OnMessage(COMM_MSG, function)
   Return r
}

Dock_ShutDown() {
	loop, 3
	   Hook_Shutdown(Dock_hHook%A_Index%, Dock_msg)
	API_FreeLibrary(Dock_hookDll)
}

Hook_Shutdown(handle, com_msg) {
   OnMessage(com_msg)
   Return DllCall("UnhookWinEvent", "UInt", handle)
}

API_ShowWindow(hwnd, flag){
   return DllCall("ShowWindow", "UInt", hwnd, "int", flag)
}

API_LoadLibrary( dll ) {
   Return DllCall("LoadLibrary", "str", dll)
}

API_FreeLibrary( h ) {
   Return DllCall("FreeLibrary", "uint", h)
}

Dock_About(pInfo="") {
	static ver := "1.0 b1"

	if pInfo = ver
		msg := ver
	else msg = 
		(LTrim
			Dock %ver%
			
			Created by majkinetor
			
			miodrag.milic@gmail.com
			code.r-moth.com
			
			Thx: Laszlo, JGR
		)

	return msg
}